# Generated by Django 3.2.23 on 2024-02-06 11:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0013_alter_serviceusageevent_question"),
    ]

    operations = [
        migrations.AddField(
            model_name="question",
            name="status",
            field=models.IntegerField(
                choices=[
                    (-100, "No status"),
                    (-3, "Failed (invalid inputs)"),
                    (-2, "Failed (timeout)"),
                    (-1, "Failed (error)"),
                    (0, "In progress"),
                    (1, "Complete"),
                ],
                default=-100,
            ),
        ),
    ]



question_basic_fieldset = (
    None,
    {
        "fields": (
            "id",
            "status",
            "service_revision",
            "asked",
            "answered",
            "latest_heartbeat",
            "duration",
        )
    },
)

question_db_input_values_fieldset = ("Input Values", {"classes": ("collapse",), "fields": ("input_values",)})
question_db_output_values_fieldset = ("Output Values", {"classes": ("collapse",), "fields": ("output_values",)})

question_delivery_ack_fieldset = (
    "Delivery Acknowledgement",
    {"classes": ("collapse",), "fields": ("delivery_acknowledgement",)},
)
question_log_records_fieldset = ("Log Records", {"classes": ("collapse",), "fields": ("log_records",)})
question_monitor_messages_fieldset = ("Monitor Messages", {"classes": ("collapse",), "fields": ("monitor_messages",)})
question_result_fieldset = ("Result", {"classes": ("collapse",), "fields": ("result",)})
question_exceptions_fieldset = ("Exceptions", {"classes": ("collapse",), "fields": ("exceptions",)})



import logging
import uuid
from datetime import datetime, timezone

from django.db import models
from django_twined.models.service_usage_events import QuestionEventsMixin
from model_utils.managers import InheritanceManager


logger = logging.getLogger(__name__)


NO_STATUS = -100
BAD_INPUT_STATUS = -3
TIMEOUT_STATUS = -2
ERROR_STATUS = -1
IN_PROGRESS_STATUS = 0
SUCCESS_STATUS = 1

STATUS_MESSAGE_MAP = {
    NO_STATUS: "No status",
    BAD_INPUT_STATUS: "Failed (invalid inputs)",
    TIMEOUT_STATUS: "Failed (timeout)",
    ERROR_STATUS: "Failed (error)",
    IN_PROGRESS_STATUS: "In progress",
    SUCCESS_STATUS: "Complete",
}

STATUS_CHOICES = tuple((k, v) for k, v in STATUS_MESSAGE_MAP.items())


class AbstractQuestion(models.Model):
    """Abstract Base Class for a Question model to store questions asked to octue services"""

    duplicate_fields = ()

    id = models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True)
    asked = models.DateTimeField(null=True, blank=True, editable=False, help_text="When the question was asked")
    answered = models.DateTimeField(null=True, blank=True, editable=False, help_text="When the question was answered")
    status = models.IntegerField(default=-100, choices=STATUS_CHOICES)

    class Meta:
        """Metaclass for AbstractQuestion"""

        abstract = True

    objects = InheritanceManager()

    def __str__(self):
        return f"{self.id}"

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.id})"

    @property
    def status_message(self):
        """Short verbose (human-readable, for display) text indicating status of the question.

        :return str:
        """
        return STATUS_MESSAGE_MAP[self.status]

    @property
    def duration(self):
        """Show the time it took to answer the question in seconds.

        :return int|None:
        """
        if self.answered and self.asked:
            return (self.answered - self.asked).seconds

    def get_duplicate(self, save=True):
        """Duplicate the question instance and optionally save to the database"""
        kwargs = {}
        for field in self.duplicate_fields:
            kwargs[field] = getattr(self, field)
        # Always override the asked and answered attributes of the duplicate
        overrides = {"asked": None, "answered": None}
        duplicate = self.__class__(**kwargs, **overrides)
        if save:
            duplicate.save()
        return duplicate

    def get_input_values(self):
        """Get the input values from wherever they're stored (or compute them)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_input_values method for this class (or provide input directly to ask() method)"
        )

    def get_input_manifest(self):
        """Get the input manifest from wherever it's stored (or compute it)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_input_manifest method for this class (or provide input directly to ask() method)"
        )

    def get_output_values(self):
        """Get the output values from wherever they're stored (or compute them) This method must be overriden in
        subclasses of Question.
        """
        raise NotImplementedError(
            "You must override the get_input_values method for this class (or provide input directly to ask() method)"
        )

    def get_output_manifest(self):
        """Get the output manifest from wherever it's stored (or compute it). This method must be overriden in
        subclasses of Question.
        """
        raise NotImplementedError(
            "You must override the get_input_manifest method for this class (or provide input directly to ask() method)"
        )

    def get_service_revision(self):
        """Get the service revision from wherever it's stored (or otherwise determine it)
        This method must be overriden in subclasses of Question
        """
        raise NotImplementedError(
            "You must override the get_service_revision method for this class (or provide input directly to ask() method)"
        )

    def as_subclass(self):
        """Returns a base question object as its inherited subclass"""
        return self.__class__.objects.get_subclass(id=self.id)

    def ask(self, save=True):
        """Ask a question to a service_revision.

        :param boolean save: If true, this question will be saved in order to update the 'asked' time.
        """

        service_revision = self.get_service_revision()
        subscription, push_url = service_revision.ask(
            question_id=self.id,
            input_values=self.get_input_values(),
            input_manifest=self.get_input_manifest(),
        )

        logger.info("Service Revision %s was asked question %s", service_revision.sruid, self.id)

        self.asked = datetime.now().replace(tzinfo=timezone.utc)

        if save:
            self.save()

        return subscription, push_url, service_revision


class Question(AbstractQuestion, QuestionEventsMixin):
    """Stores questions asked to octue services

    This concrete model is here to link questions to the service revisions
    used to ask them. It is intended to be subclassed for specific question types,
    hence why it doesn't override the getters for input and output values.

    """

    service_revision = models.ForeignKey(
        "django_twined.ServiceRevision",
        blank=True,
        null=True,
        related_name="questions",
        on_delete=models.PROTECT,
    )

    def get_service_revision(self):
        return self.service_revision



from .datastores import AbstractSynchronisedDatastore
from .questions import (
    BAD_INPUT_STATUS,
    ERROR_STATUS,
    IN_PROGRESS_STATUS,
    NO_STATUS,
    SUCCESS_STATUS,
    TIMEOUT_STATUS,
    AbstractQuestion,
    Question,
)
from .service_revisions import AbstractServiceRevision, ServiceRevision, get_default_service_revision
from .service_usage_events import QUESTION_ASKED, QUESTION_RESPONSE_UPDATED, QUESTION_STATUS_UPDATED, ServiceUsageEvent


__all__ = (
    "AbstractServiceRevision",
    "AbstractSynchronisedDatastore",
    "AbstractQuestion",
    "get_default_service_revision",
    "Question",
    "NO_STATUS",
    "BAD_INPUT_STATUS",
    "TIMEOUT_STATUS",
    "ERROR_STATUS",
    "IN_PROGRESS_STATUS",
    "SUCCESS_STATUS",
    "QUESTION_ASKED",
    "QUESTION_RESPONSE_UPDATED",
    "QUESTION_STATUS_UPDATED",
    "ServiceRevision",
    "ServiceUsageEvent",
)



import logging

from django.db import models
from django.db.models import Q


logger = logging.getLogger(__name__)


QUESTION_ASKED = "q-asked"
QUESTION_RESPONSE_UPDATED = "q-response-updated"
QUESTION_STATUS_UPDATED = "q-status-updated"

SERVICE_USAGE_EVENT_KINDS = {
    QUESTION_ASKED: "Question asked",
    QUESTION_RESPONSE_UPDATED: "Question response updated",
    QUESTION_STATUS_UPDATED: "Question status updated",
}

SERVICE_USAGE_EVENT_KINDS_CHOICES = tuple((k, v) for k, v in SERVICE_USAGE_EVENT_KINDS.items())


class AbstractEvent(models.Model):
    """Abstract base model for storing events

    This allows creation of multiple different tables for storage of different events.

    """

    id = models.BigAutoField(primary_key=True)

    data = models.JSONField(blank=False, null=False, editable=False, help_text="Event payload")

    kind = models.CharField(
        max_length=20,
        null=False,
        blank=False,
        help_text="Event kind",
        choices=SERVICE_USAGE_EVENT_KINDS_CHOICES,
    )

    publish_time = models.DateTimeField(null=False, blank=False, help_text="Event timestamp")

    class Meta:
        """AbstractEvent meta class data"""

        abstract = True


class ServiceUsageEvent(AbstractEvent):
    """Table for storing Service Usage Events

    This table will store events related to usage of octue services (for example, question asks, answers, updates and responses)
    in your main database.

    For high volume processing, you may wish not to store events at all, or you may wish to store in a different
    database such as BigQuery. More helper tools for this coming soon.

    This table relates to concrete Question and Service Revision models. Null relations are
    not allowed, meaning that the table cannot store events not generated with django-twined
    (i.e. where Service Revision and Question models do not exist). This may change in future,
    to enable absorption of questions triggered by other services.
    """

    question = models.ForeignKey(
        "django_twined.Question",
        null=False,
        blank=False,
        related_name="service_usage_events",
        on_delete=models.CASCADE,
        editable=False,
    )

    service_revision = models.ForeignKey(
        "django_twined.ServiceRevision",
        null=False,
        blank=False,
        related_name="service_usage_events",
        on_delete=models.PROTECT,
        editable=False,
    )

    def __str__(self):
        return str(self.kind)

    def __repr__(self):
        return f"Service Usage Event {self.kind}"


class QuestionEventsMixin:
    """A mixin for the `Question` subclass providing helpers for retrieval of specific message kinds. These methods are
    backwards compatible with database entries created before the breaking change in version `0.7.0` was introduced,
    allowing the new and old formats of JSON data stored in the events to be accessed.
    """

    @property
    def delivery_acknowledgement(self):
        """Get the delivery acknowledgement for the question.

        :return django_twined.models.querysets.datastore_queryset.DatastoreQueryset:
        """
        try:
            return self.service_usage_events.get(self._get_event_filter("delivery_acknowledgement"))
        except ServiceUsageEvent.DoesNotExist:
            return None
        except ServiceUsageEvent.MultipleObjectsReturned:
            logger.warning(
                "MultipleObjectsReturned detected for delivery_acknowledgement ServiceUsageEvent on question %s",
                self.id,
            )
            return self.service_usage_events.filter(self._get_event_filter("delivery_acknowledgement")).first()

    @property
    def exceptions(self):
        """Get any exceptions raised by the child service during processing of the question.

        :return django_twined.models.querysets.datastore_queryset.DatastoreQueryset:
        """
        return self.service_usage_events.order_by("publish_time").filter(self._get_event_filter("exception")).all()

    @property
    def result(self):
        """Get the result produced by the child service in response to the question.

        :return django_twined.models.querysets.datastore_queryset.DatastoreQueryset:
        """
        try:
            return self.service_usage_events.get(self._get_event_filter("result"))
        except ServiceUsageEvent.DoesNotExist:
            return None
        except ServiceUsageEvent.MultipleObjectsReturned:
            logger.warning("MultipleObjectsReturned detected for result ServiceUsageEvent on question %s", self.id)
            return self.service_usage_events.filter(self._get_event_filter("result")).first()

    @property
    def log_records(self):
        """Get any log records produced by the child service processing the question.

        :return django_twined.models.querysets.datastore_queryset.DatastoreQueryset:
        """
        return self.service_usage_events.order_by("publish_time").filter(self._get_event_filter("log_record")).all()

    @property
    def monitor_messages(self):
        """Get any monitor messages produced by the child service processing the question.

        :return django_twined.models.querysets.datastore_queryset.DatastoreQueryset:
        """
        return (
            self.service_usage_events.order_by("publish_time").filter(self._get_event_filter("monitor_message")).all()
        )

    @property
    def latest_heartbeat(self):
        """Get the latest heartbeat of the child service processing the question.

        :return django_twined.models.querysets.datastore_queryset.DatastoreQueryset:
        """
        return self.service_usage_events.order_by("-publish_time").filter(self._get_event_filter("heartbeat")).first()

    def _get_event_filter(self, data_type_or_kind):
        """Get a filter for `ServiceUsageEvent` model instances that filters the JSON data of the `data` field for the
        given event kind using either the old `type` key or the new `kind` key. This maintains backwards compatibility
        with service usage events created before the breaking change in version `0.7.0` was introduced, allowing the new
        and old formats of the JSON data stored in the events to be accessed.

        :param str data_type_or_kind: the name of the event type/kind to filter for
        :return django.db.models.query_utils.Q:
        """
        return Q(data__type=data_type_or_kind) | Q(data__kind=data_type_or_kind)
